/*
	Copyright (C) 2011 Micha Hersch, University of Lausanne, Swiss Institute of Bioinformatics

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/*!
 *	\file openCotyledon.cpp
 *	\author Micha Hersch
 *	\date
 */
#include "openCotyledon.h"

#define MAX_STEP 20

using namespace std;

int OpenCotyledon::of_displ = 5;

OpenCotyledon::OpenCotyledon(){
  maxFeatures = 100;
  pre_features = new Vec2[maxFeatures];
  post_features = new Vec2[maxFeatures];
  of_feature_found = new char[maxFeatures];
  of_feature_error = new f32[maxFeatures];
  position = Vec2(-1,-1);
  active = 0;
  win_size =Vec2(201,201);
}

int OpenCotyledon::computeOF(IplImage *img, IplImage *img_old){
  CvSize siz = cvGetSize(img);
  IplImage *pyr1 = cvCreateImage(siz,IPL_DEPTH_8U,1);
  IplImage *pyr2 = cvCreateImage(siz,IPL_DEPTH_8U,1);
  IplImage *eig = cvCreateImage(siz,IPL_DEPTH_32F,1);
  IplImage *tmp = cvCreateImage(siz,IPL_DEPTH_32F,1);
  
  for(int i=0;i<maxFeatures;i++){
	  post_features[i] = Vec2(-1,-1);
	  of_feature_found[i] = 0;
  }
  int nbFeatures=maxFeatures;
   float quality=0.1;
  do{
    nbFeatures=maxFeatures;
    //   cout<<"features "<<nbFeatures<< endl;  
    cvGoodFeaturesToTrack(img_old, eig, tmp, pre_features, &nbFeatures, quality, 2, NULL,3,0);
    
    quality *=0.5;
    //    cout<<"nb features "<<nbFeatures<<endl;

  }while(nbFeatures<20 && quality>0.000001);
 
  if(nbFeatures<5){
    cerr<<"to few features "<<nbFeatures<< endl;
    //  return;n
  }
  else{  
    CvSize of_window = cvSize(2*of_displ+1,2*of_displ+1);
    CvTermCriteria criteria = cvTermCriteria( CV_TERMCRIT_ITER | CV_TERMCRIT_EPS, 20, .3 );
    
    cvCalcOpticalFlowPyrLK(img_old,img, pyr1, pyr2, pre_features, 
			   post_features, nbFeatures, of_window, 3,
			   of_feature_found, of_feature_error, criteria,0);
    
    for(int i=0;i<nbFeatures;i++){
      if(of_feature_found[i]){
	Vec2 pf = post_features[i];
	if(pf.x<=-1 || pf.y<=-1 ||
	   (pf-pre_features[i]).lengthSquared()>MAX_STEP*MAX_STEP){
	  of_feature_found[i] = 0;
	}
      }
    }
  }
  return nbFeatures;
}


Vec2 OpenCotyledon::update(IplImage *img,IplImage *img_old){

  int nx,ny,lx,ly,wx,wy;
  wx = win_size.x/2;
  wy = win_size.y/2;   
  nx = max(0,position.x-wx);
  ny = max(0,position.y-wy);
  lx = min(position.x+wx-nx+1,img->width-nx);
  ly = min(position.y+wy-ny+1,img->height-ny);
  CvRect roi= cvRect(nx,ny,lx,ly);
  Vec2 lpos = position-Vec2(nx,ny); //position in ROI FoR
  cvSetImageROI(img,roi);
  cvSetImageROI(img_old,roi);
  //  cout<<Vec2(nx,ny)<<" "<<Vec2(lx,ly)<<endl;
  int nbF=computeOF(img, img_old);
  cvResetImageROI(img);
  cvResetImageROI(img_old);
 
  int n = 0;
  for(int i=0;i<nbF;i++){
     if(of_feature_found[i]){
       n++;
     }
   }
  if(n<5){
    cerr<<"Only found "<<n<<" features"<<endl;
    cout<<position<<endl;
    return Vec2(0,0);
  }
  //  cout<<n<<" out of "<<nbF<<endl;
  // cvWaitKey();

  CvMat *diff = cvCreateMat(n,2,CV_32FC1);
  CvMat *difft = cvCreateMat(2,n,CV_32FC1);
  CvMat *k = cvCreateMat(n,1,CV_32FC1);
  CvMat *projk = cvCreateMat(n,1,CV_32FC1);
  int cnt= 0;
  for(int i=0;i<nbF;i++){
    if(of_feature_found[i]){
      cvSet2D(diff,cnt,0,cvScalar(pre_features[i].x - lpos.x));
      cvSet2D(diff,cnt,1,cvScalar(pre_features[i].y - lpos.y)); 
      cvSet1D(k,cnt,cvScalar(100/((float)(pre_features[i]-lpos).lengthSquared())));// distance to cotyledon
      cnt++;
    }
  }
  
  CvMat *proj = cvCreateMat(n,n,CV_32FC1);
  CvMat *m1 = cvCreateMat(2,2,CV_32FC1);
  CvMat *m2 = cvCreateMat(2,2,CV_32FC1);
  CvMat *m3 = cvCreateMat(n,2,CV_32FC1);
  CvMat *m4 = cvCreateMat(n,n,CV_32FC1);
  CvMat *m5 = cvCreateMat(n,n,CV_32FC1);
  cvTranspose(diff,difft);
  cvMulTransposed(diff,m1,1);
  cvInv(m1,m2);
  cvMatMul(diff,m2,m3);
  cvMatMul(m3,difft,m4);
  cvSetIdentity(m5);
  cvSub(m5,m4,proj);
  cvMatMul(proj,k,projk);
  float den=0;
  Vec2 num(0,0);
  cnt=0;
   for(int i=0;i<nbF;i++){
     if(of_feature_found[i]){
       float a= cvGet1D(projk,cnt).val[0];
       num += post_features[i]*a;
       den+=a;
       cnt++;
     }
   }
   Vec2 npos =  num/den;
   if(npos.x<0 || npos.y<0){
     cout<<"bad position "<<npos<<endl;
     return Vec2(0,0);
   }
   else{
     position = npos+Vec2(nx,ny);
     return npos-lpos;
   }
}

void OpenCotyledon::displayFeatures(IplImage *img,int n,const char *wname){
  IplImage *dim = cvCreateImage(cvGetSize(img),8,3);
  cvMerge(img,img,img,NULL,dim);
  for(int i=0;i<n;i++){
    CvScalar col = of_feature_found[i]?CV_RGB(255,0,0):CV_RGB(100,0,0);
    //CvScalar col =cvScalar(100);
    cvRectangle(dim,point2(pre_features[i]-Vec2(1,1)),
		point2(pre_features[i]+Vec2(1,1)),col,-1);
  }
  cvShowImage(wname,dim);
  IMKILL(dim);
}

void OpenCotyledon::displayMatchingFeatures(IplImage *img,int n,const char *wname){
  IplImage *dim = cvCreateImage(cvGetSize(img),8,3);
 cvMerge(img,img,img,NULL,dim); 
 for(int i=0;i<n;i++){
    if(of_feature_found[i]){
      // cvRectangle(tuner.img,point2(post_features[i]-Vec2(1,1)),
//  		  point2(post_features[i]+Vec2(1,1)),cvScalar(100),-1);
      cvLine(dim,point2(pre_features[i]),point2(post_features[i]),
	     CV_RGB(255,0,0)); //   cvScalar(100));
    }
  }
  if(position.x>0&&position.y>0){
    //    cvRectangle(img,point2(position-Vec2(1,1)),point2(position+Vec2(1,1)),
    //		CV_RGB(0,0,255),-1);
  }
  cvShowImage(wname,dim);
  //  cvWaitKey();
  IMKILL(dim);
}




#ifdef OPEN_COTYLEDON_DEV
void OpenCotyledon::runTuner(const char *imfname){
  tuner.init(imfname);
  cvSetMouseCallback("cot2",on_mouse,this);
  tuner.nextFrame();
  //  waitForCenter();
  do{
    update();
    tuner.display();
    while(' '!= cvWaitKey()){}
  }while(tuner.nextFrame());  
}

#endif

// used for development purposes
#define OPEN_COTYLEDON_MAIN
#ifdef OPEN_COTYLEDON_DEV
#ifdef OPEN_COTYLEDON_MAIN
int main(int argc, char *argv[]){
  OpenCotyledon cot;

  if(argc<1){return 0;}
  cot.runTuner(argv[1]);
  return 1;
}
#endif
#endif
